import os
from enum import Enum
from typing import List, Dict, Optional

import yaml
from pydantic import BaseModel, PositiveInt


DEFAULT_CONFIG_PATHS = [
    os.path.expanduser(os.path.join("~", ".config", "mediasorter.yml")),
    os.path.join("etc", "mediasorter", "mediasorter.yml")
]


class MediaType(Enum):
    TV_SHOW = "tv"
    MOVIE = "movie"
    AUTO = "auto"


class Action(Enum):
    MOVE = "move"
    HARDLINK = "hardlink"
    SYMLINK = "symlink"
    COPY = "copy"


class MetadataProviderApi(BaseModel):
    name: str
    key: Optional[str]
    url: Optional[str]
    path: Optional[str]


class OperationOptions(BaseModel):
    user: str = "root"
    group: str = "media"
    chown: bool = False
    dir_mode: str = '0o644'
    file_mode: Optional[str]
    overwrite: bool = False
    infofile: bool = False
    shasum: bool = False


class ScanConfig(BaseModel):
    """
    Specify an "input/output" combo for different directories
    """
    src_path: str  # source path (duh...)
    media_type: MediaType = MediaType.AUTO  # force only a specific media type
    action: Action = Action.COPY  # select action type
    tv_shows_output: Optional[str]  # where to put recognized TV shows
    movies_output: Optional[str]  # where to put recognized movies
    options: OperationOptions = OperationOptions()  # options for the sorting operation itself


class BaseParams(BaseModel):
    min_split_length: PositiveInt = 3
    suffix_the: bool = False
    search_overrides: Dict[str, str] = {}
    name_overrides: Dict[str, str] = {}
    file_format: str


class MovieParams(BaseParams):
    subdir: bool = True  # sort all files related to a single movie to a common subdir
    file_format: str = "{title} ({year})"
    dir_format: str = file_format
    allow_metadata_tagging: bool = False


class TvShowParams(BaseParams):
    dir_format: str = "{series_title}/Season {season_id}"
    file_format: str = '{series_title} - S{season_id:02d}E{episode_id:02d} - {episode_title}'


class Parameters(BaseModel):
    valid_extensions: List[str] = [".avi", ".mkv", ".mp4"]
    split_characters: List[str] = [" ", ".", "_"]

    tv: TvShowParams = TvShowParams()
    movie: MovieParams = MovieParams()


class Logging(BaseModel):
    logfile: str
    loglevel: str


class MediaSorterConfig(BaseModel):
    # Configure different metadata provider APIs (API keys, override URLs,...).
    # Must correspond to an existing key in the MetadataProvider enum.
    api: List[MetadataProviderApi] = []

    # Configure multiple directories to be scanned
    # without the need to specify using command line interface.
    scan_sources: Optional[List[ScanConfig]] = None

    parameters: Parameters = Parameters()

    metainfo_map: Dict[str, str] = {}

    loging: Optional[Logging]


def read_config(config_file: str) -> MediaSorterConfig:

    with open(config_file, 'r') as cfgfile:
        o_config = yaml.load(cfgfile, Loader=yaml.SafeLoader)

    return MediaSorterConfig(**o_config['mediasorter'])



import logging
import os.path
import shutil
import subprocess
from subprocess import DEVNULL


from mediasorter.config import Action


log = logging.getLogger(".".join([__package__, __name__]))


class ExecutionError(Exception):
    pass


class Executable:

    @classmethod
    def from_action_type(cls, action: Action):
        if action == Action.COPY:
            return Copy()
        if action == Action.MOVE:
            return Move()
        if action == Action.SYMLINK:
            return RunSubprocess('ln', '-s', stdout=DEVNULL, stderr=DEVNULL)
        if action == Action.HARDLINK:
            return RunSubprocess('ln', stdout=DEVNULL, stderr=DEVNULL)
        raise NotImplementedError(f"Action executor '{action}' not implemented")

    def _commit(self, source, destination):
        pass

    def __str__(self) -> str:
        return f"{__class__.__name__}"

    def commit(self, source, destination):
        log.info(f"[{self}: sorting: {source} -> {destination}")
        try:
            # Sanity check
            if not os.path.exists(source):
                raise FileNotFoundError(f"File not found '{source}'")

            # Create parent dir(s)
            parent_dir = os.path.dirname(destination)
            if not os.path.isdir(parent_dir):
                log.info(f"Creating target directory '{parent_dir}'")
                os.makedirs(parent_dir)

            # Let's go
            return self._commit(source, destination)
        except Exception as e:
            raise ExecutionError(e)


class RunSubprocess(Executable):
    def __init__(self, *args, **kwargs) -> None:
        self.args = args
        self.kwargs = kwargs

    def __str__(self) -> str:
        return f"{super}(args={self.args})"

    def _commit(self, source, destination):
        args = self.args + (source, destination)
        output = subprocess.run(args, **self.kwargs)
        if output.returncode != 0:
            raise ExecutionError(f"'{args}' subprocess failed. {output.stdout=}, {output.stderr=}")


class Copy(Executable):
    def _commit(self, source, target):
        return shutil.copy(source, target)


class Move(Executable):
    def _commit(self, source, target):
        return shutil.move(source, target)


import asyncio
import logging
import random
import re
from enum import Enum
from typing import Dict, Optional, Tuple, Callable, Any
from urllib.parse import quote

import aiohttp
from aiohttp import ClientResponseError
from cache import AsyncLRU
from pydantic import BaseModel

from .config import MetadataProviderApi
from .utils import split_and_lower

log = logging.getLogger(".".join([__package__, __name__]))


class MetadataQueryError(Exception):
    pass


class MetadataNotFoundError(MetadataQueryError):

    search_term: str


class TvShowMetadata(BaseModel):
    series_title: str
    season_id: int
    episode_title: str
    episode_id: int


class MovieMetadata(BaseModel):
    title: str
    year: int


def validation(fn):
    # just a dummy marker
    return fn


class MetadataApi:

    # These attributes can be hard-coded in the implementation
    # and overriden via configuration.
    key: str = None
    url: str = None
    path: str = None

    def __init__(self, config: MetadataProviderApi = None) -> None:
        if config and config.key:
            self.key = config.key
        if config and config.url:
            self.url = config.url
        if config and config.path:
            self.path = config.path

    async def query(self, *args, **kwargs):
        """
        The high level query function used to query the database. Returns media metadata
        in a format specific to the actual implementation.

        MUST return a valid response, otherwise raise any MetadataQueryException.
        """
        raise NotImplemented

    @staticmethod
    @AsyncLRU(maxsize=1024)
    async def request(url: str) -> dict:
        """
        A wrapper around the actual request that caches a successful response is obtained.
        Otherwise, ClientResponseError is raised.

        :param url: Specify the url that we want to request
        :return: A json object
        """
        log.debug(f"Network request: [GET]({url})")
        async with aiohttp.ClientSession() as session:
            async with session.get(url) as response:
                response.raise_for_status()  # Raised errors don't get cached.
                data = await response.json()
        return data

    async def async_fetch_json(self, url, retry=0, max_retries=4):
        """
        A wrapper around the requests function that handles retries and error handling.
        It returns the raw JSON response, or raises an exception if it eventually fails
        to get a valid response.

        :param url: specify the url to fetch
        :param retry: int: retry counter (0 = the actual 1st request, not a re-try yet)
        :param max_retries: int: limit the number of retries in case of a 429 or 503 errors
        :return: the result of the metadata API
        """
        log.debug(f"fetching '{url}', retries={retry}/{max_retries}")
        try:
            return await MetadataApi.request(url)
        except ClientResponseError as e:
            # Running too many queries at once causes the API (TV Maze at the very least)
            # to decline our requests. Simple wait and retry seems to be sufficient.
            if e.status in (429, 503) and retry < max_retries:
                wait_sec = random.randint(10, 100) / 100 * (retry + 1)
                log.warning(f"Service unavailable, retrying in {wait_sec}s.")
                await asyncio.sleep(wait_sec)
                return await self.async_fetch_json(url, retry + 1)
            raise MetadataQueryError(e)

    async def try_harder(
        self,
        search_term: str,
        validation_mapper: Callable[[dict, Any], Any] = None,  # must raise
        validation_callback_args: Tuple = None,
        to_be_raised: Exception = None,
        try_index: int = 1,
        min_len: int = 1
    ) -> Any:
        """
        A helper function that tries to find the metadata by searching for it in an iterative way.
        It does this by removing one word from the end of the search term and trying again. If it
        fails, then it removes another word and tries again until min_len is reached.

        :param search_term: Search for the show
        :param validation_mapper: Validate the response data (! must raise MetadataQueryException)
        :param validation_callback_args:Tuple=None: Pass arguments to the validation callback
        :param to_be_raised: Pass an exception to be raised once the function eventually fails.
        :param try_index: Keep track of the number of tries
        :param min_len: set the minimum words count to be searched for.
        :return: The response data "as-is" if it matches provided validation fn.
        """
        split = search_term.split()

        while len(split) >= min_len:
            search_title = self.clean_search_term(" ".join(split))
            # search_title = quote(search_title_in)
            log.info(f"{try_index}. try: search='{search_title}'")
            show_url = self.url + "/" + self.path.format(title=quote(search_title))
            try:
                response_data = await self.async_fetch_json(show_url)
                return await validation_mapper(response_data, search_title, *(validation_callback_args or tuple()))
            except MetadataQueryError as e:
                log.debug(f"Invalid search result: {e}")
                if not to_be_raised:
                    to_be_raised = e
                return await self.try_harder(
                    search_term=" ".join(split[:-1]),
                    validation_mapper=validation_mapper,
                    validation_callback_args=validation_callback_args,
                    to_be_raised=to_be_raised,  # propagate "the original" exc
                    try_index=try_index + 1,
                    min_len=min_len
                )
        raise to_be_raised or MetadataQueryError(f"No exception provided, query failed: "
                                                 f"{search_term=}, {try_index}. try.")

    def clean_search_term(self, string, overrides: Dict[str, str] = None) -> str:
        parts = string.split()
        if not parts:
            parts = [string]

        # Skip all the words "the" in the title, because TVMaze seems to choke on this.
        # * The leading 'the' needs to be preserved though...
        first_the = None
        if re.match('^[Tt]he$', parts[0]):
            first_the = parts.pop(0)
        parts = [word for word in parts if not re.match('^[Tt]he$', word)]
        if first_the:
            parts.insert(0, first_the)

        # rejoin
        # search_title = ' '.join(parts)  # This seems to yield better results than '+'.
        search_title = ' '.join(parts)  # This seems to yield better results than '+'.

        # search_title = title
        if overrides and search_title in overrides:
            new_name = overrides[string]
            log.debug(f"Overriding search title: '{string}' -> '{new_name}'")
            search_title = new_name

        search_title = search_title.replace("'", "")

        return search_title


class TvShowMetadataApi(MetadataApi):

    async def query(self, title: str, season_id: int, episode_id: int) -> TvShowMetadata:
        raise NotImplemented


class MovieMetadataApi(MetadataApi):

    async def query(self, title: str, year: str) -> MovieMetadata:
        raise NotImplemented


class TvMaze(TvShowMetadataApi):

    # Base URL for TV Maze; should generally not be changed (can be overriden via config)
    url: str = "https://api.tvmaze.com"

    # Search path for TV Maze; should generally not be changed (can be overriden via config)
    # Variables replaced at runtime:
    #  * "title": the show name query, extracted from the source filename
    path = "singlesearch/shows?q={title}&embed=episodes"

    async def _match_tv_show(
            self, series_data: dict, search_title: str, season_id: int, episode_id: int
    ) -> Tuple[dict, dict]:
        """
        Find the correct episode in the series episodes list.

        :param series_data: dict: Pass the data from _get_series_data() to _match_tv_show()
        :param season_id: int: Specify the season number
        :param episode_id: int: Specify the episode number
        :return: A dictionary with the episode data
        """

        @validation
        def find_episode(e_list):
            for episode in e_list:
                if episode.get('season') == season_id and episode.get('number') == episode_id:
                    return episode

        episode_list = series_data['_embedded'].get('episodes', [])
        correct_episode = find_episode(episode_list)
        if not correct_episode:
            log.warning(f"Episode {season_id=} {episode_id=} not found.")

            if series_data.get("_links") and series_data["_links"].get("self"):
                alternative_url = series_data["_links"]["self"]["href"] + "/episodes?specials=True"
                log.debug(f"Trying alternative URL: {alternative_url}")

                episode_list = await self.async_fetch_json(alternative_url)

                correct_episode = find_episode(episode_list)
                if not correct_episode:
                    # One last try - maybe TV Maze does not show the episode id,
                    # e.g. a special episode (?)
                    season_only = list(filter(lambda e: e["season"] == season_id, episode_list))
                    season_only.sort(key=lambda e: e["airdate"])
                    log.info(f"{len(season_only)} {episode_id=}")
                    episode_list = season_only
                    if len(season_only) >= episode_id:
                        # Yay! We should have a match. Just grab the episode by index
                        # (assuming the 'air date' ordering is valid).
                        correct_episode = season_only[episode_id - 1]

        if not correct_episode:
            episodes = [f"S{e.get('season')}E{e.get('number')}" for e in episode_list]
            msg = f"TV show {series_data.get('name')} found, BUT correct " \
                  f"episode {season_id=}, {episode_id=} can't be found in {episodes}"
            raise MetadataQueryError(msg)
        return series_data, correct_episode

    async def query(self, title: str, season_id: int, episode_id: int) -> TvShowMetadata:

        series_data, episode = await self.try_harder(
            search_term=title,
            validation_mapper=self._match_tv_show,
            validation_callback_args=(season_id, episode_id)
        )

        # Sometimes, we get a slash; only invalid char on *NIX so replace it
        episode_title = episode.get('name').replace('/', '-')

        return TvShowMetadata(
            series_title=series_data.get('name'),
            season_id=season_id,
            episode_title=episode_title,
            episode_id=episode_id
        )


class TMDB(MovieMetadataApi):

    # Base URL for TMDB; should generally not be changed (can be overriden via config)
    url: str = "https://api.themoviedb.org/3"

    # Search path for TMDB; should generally not be changed (can be overriden via config)
    # Variables replaced at runtime:
    #  * "key": the TMDB API key specified above
    #  * "title": the movie name query, extracted from the source filename
    path: str = "search/movie?api_key={key}&query={title}"

    # Don't load more than this amount of pages - could get pretty crazy.
    max_pages = 5

    async def async_fetch_json(self, url, retry=0, max_retries=4):
        """
        Override for TMDB in case there is more than 1 page of results.
        Very generic search terms can produce A LOT of results,
        but our target year might still be somewhere in there...
        """
        result = await super().async_fetch_json(url, retry, max_retries)
        total_pages = result.get("total_pages", 0)
        if total_pages <= 1 or not result.get("results"):
            return result

        next_page = 2
        while next_page <= (self.max_pages if total_pages >= self.max_pages else total_pages):
            next_page_data = await super().async_fetch_json(url + f"&page={next_page}")
            next_page += 1
            if next_results := next_page_data.get("results"):
                result.get("results").extend(next_results)

        return result

    def __init__(self, config: MetadataProviderApi = None) -> None:
        super().__init__(config)
        self.path = self.path.format(key=self.key, title='{title}')

    def clean_search_term(self, string, overrides: Dict[str, str] = None):
        # "Sanitize" input name...

        # Remove the first "The" from the title when searching to avoid weird conflicts
        search_movie_title = re.sub("[Tt]he\+", "", string, 1)
        search_movie_title = search_movie_title.replace("'", "")
        # Apply overrides
        if overrides and (search_movie_title in overrides):
            new_name = overrides[search_movie_title]
            log.debug(f"Overriding search title: '{search_movie_title}' -> '{new_name}'")
            search_movie_title = new_name

        return search_movie_title

    @staticmethod
    def _parse_release_year(movie_data: dict):
        try:
            return int(movie_data.get('release_date', '0000-00-00').split('-')[0])
        except (TypeError, ValueError):
            return 0  # No release year.

    @validation
    async def _match_movie(self, movie_data: dict, search_term, search_year: int = None):
        # List all movies and find the one with matching release year (+- 1 year)
        result_list = movie_data.get('results')

        if not result_list:
            raise MetadataQueryError(f"'{search_term}': Results list is empty.")

        new_result_list = result_list
        if search_year:
            new_result_list = list(
                filter(
                    lambda res: abs(self._parse_release_year(res) - search_year) < 2, result_list
                )
            )
        elif len(result_list) > 2:  # We need to be careful, we want to fail here.
            raise MetadataQueryError(
                f"{search_term}: too many results ({len(result_list)}), we cant' be sure"
            )

        if not new_result_list:
            raise MetadataQueryError(
                f"Movie search produced results, BUT the requested {search_year=} not found in: "
                f"{['{title}/{release_date}'.format(**mov) for mov in result_list]}"
            )
        probable_result = new_result_list[0]

        result_movie_title = probable_result.get('title')
        result_original_movie_title = probable_result.get('original_title')
        result_movie_year = self._parse_release_year(probable_result)

        # Triple check: flimsy, but better than nothing...
        # Look for at least a single matching word in the result title(s).
        search_terms = split_and_lower(search_term, alphanum_only=True)
        actual_terms = split_and_lower(result_original_movie_title, alphanum_only=True)
        actual_terms = actual_terms + split_and_lower(result_movie_title, alphanum_only=True)
        if not any([t in actual_terms for t in search_terms]):
            raise MetadataQueryError(
                f"'{search_term}': result '{result_movie_title}' probably a nonsense."
            )

        return result_movie_title, result_movie_year

    async def query(self, search_movie_title: str, search_year: int) -> Optional[MovieMetadata]:
        if not self.key:
            raise MetadataQueryError(f"{self.__class__.__name__}: key required.")

        result_movie_title, result_movie_year = await self.try_harder(
            search_term=search_movie_title,
            validation_mapper=self._match_movie,
            validation_callback_args=(search_year,)
        )
        return MovieMetadata(title=result_movie_title, year=result_movie_year)


class MetadataProvider(Enum):
    """List of all available (implemented) metadata providers."""
    TV_MAZE = "tvmaze"
    TMDB = "tmdb"

    @property
    def clazz(self):
        return {
            MetadataProvider.TV_MAZE.value: TvMaze,
            MetadataProvider.TMDB.value: TMDB,
        }.get(self.value)



import re
from typing import List


def replace_non_alphanum(string: str) -> str:
    return re.sub('[^0-9a-zA-Z]+', '', string)


def split_and_lower(string: str, alphanum_only=False) -> List[str]:
    split = string.lower().split()
    if alphanum_only:
        return [replace_non_alphanum(word) for word in split]
    return split



import asyncio
import logging
import os
import re
import sys
from shutil import copyfile
from typing import Optional

import click
import pkg_resources
from pydantic import ValidationError
from rich.console import Console
from rich.progress import Progress, BarColumn, TextColumn
from rich.prompt import Confirm, Prompt
from rich.table import Column
from rich.text import Text

from mediasorter.config import MediaSorterConfig, read_config, DEFAULT_CONFIG_PATHS, ScanConfig, \
    OperationOptions
from mediasorter.sorter import MediaSorter, Action, MediaType

FORMAT = "%(asctime)s [%(levelname)s] %(message)s"
logging.basicConfig(level="NOTSET", format=FORMAT)
logging.getLogger('asyncio').setLevel(logging.WARNING)

log = logging.getLogger()

# Global Click options
CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'], max_content_width=120)


def _pretty_print_operation(sort_operation, console, before=False):
    if before and not sort_operation.is_error:
        console.print(Text(f"○ {sort_operation.input_path}", style="green"))
        console.print(Text(f"   ⤷  [{sort_operation.action.value}] {sort_operation.output_path}", style="green"))
    elif not sort_operation.is_error:
        console.print(Text(f" ✓ {sort_operation.output_path}", style="green"))

    elif sort_operation.is_error:
        console.print(Text(f"○ {sort_operation.input_path}", style="red"))
        console.print(Text(f"   ⤬ {sort_operation.exception}", style="#fe9797"))
    else:
        console.print(Text(f"⚠ {sort_operation.input_path}", style="yellow bold"))


def _load_config(console, config_files, verbose=False) -> MediaSorterConfig:
    """Load first valid configuration file."""
    lines = []
    exception: Optional[Exception] = None
    for cfg_path in config_files:
        try:
            cfg = read_config(cfg_path)
            log.info(f"Using configuration file '{cfg_path}'")
            return cfg
        except FileNotFoundError:
            lines.append(
                Text(f"Can't load configuration '{cfg_path}', file not found", style="yellow")
            )
        except (KeyError, ValidationError) as e:
            lines.append(
                Text(
                    f"Can't load configuration '{cfg_path}', invalid config keys.",
                    style="yellow"
                )
            )
            exception = e
        except Exception as e:
            lines.append(
                Text(
                    f"Can't load configuration from '{cfg_path}', unexpected error {e} ",
                    style="yellow"
                )
            )
            exception = e
    for line in lines:
        console.print(line)
    if verbose:
        raise exception


@click.command(
    context_settings=CONTEXT_SETTINGS,
    help="Scan and sort one or more source files/directories (directories will be crawled"
         "recursively). Media sources (and respective destination paths) can also be specified "
         "using the configuration file.\n\n"
         "Metadata will be searched for using configured APIs and media file names"
         "will be renamed accordingly."
)
@click.option(
    '-d', '--destination', 'dst_path',
    type=click.Path(),
    default='~/Media', show_default=True,
    help='The target directory to sort to'
)
@click.option(
    '-dtv', '--destination-tv', 'dst_path_tv',
    type=click.Path(),
    default='~/Media', show_default=True,
    help='The target directory to sort to: TV shows ONLY, shadows the --destination option'
)
@click.option(
    '-dmov', '--destination-mov', 'dst_path_mov',
    type=click.Path(),
    default='~/Media', show_default=True,
    help='The target directory to sort to: movies ONLY, shadows the --destination option'
)
@click.option(
    '-t', '--type', 'mediatype',
    type=click.Choice([t.value for t in MediaType]),
    default='auto', show_default=True,
    help='The type of media to aid the sorter'
)
@click.option(
    '-a', '--action', 'action',
    type=click.Choice([a.value for a in Action]),
    default='copy', show_default=True,
    help='How to get the media to the destination'
)
@click.option(
    '-i/-I', '--infofile/--no-infofile', 'infofile',
    is_flag=True, default=False, show_default=True,
    help="Create information file at target."
)
@click.option(
    '-s/-S', '--shasum/--no-shasum', 'shasum',
    is_flag=True, default=False, show_default=True,
    help="Create SHA256sum file at target."
)
@click.option(
    '-o/-O', '--chown/--no-chown', 'chown',
    is_flag=True, default=False, show_default=True,
    help="Change ownership and permissions of destfile to match user/group and mode."
)
@click.option(
    '-u', '--user', 'user',
    default='root', show_default=True,
    help='The user that should own the sorted files if --chown'
)
@click.option(
    '-g', '--group', 'group',
    default='media', show_default=True,
    help='The group that should own the sorted files if --chown'
)
@click.option(
    '-mf', '--file-mode', 'file_mode',
    default='0o644', show_default=True,
    help='The Python-octal-format permissions for the target files if --chown'
)
@click.option(
    '-md', '--directory-mode', 'directory_mode',
    default='0o755', show_default=True,
    help='The Python-octal-format permissions for the created file parent directory if --chown'
)
@click.option(
    '-tm', '--tag-metainfo', 'metainfo_tag',
    is_flag=True, default=False, show_default=True,
    help="Add metainfo tagging to target filenames (see README)."
)
@click.option(
    '-o', '--overwrite', 'overwrite',
    is_flag=True, default=False, show_default=True,
    help='Replace files at the destination'
)
@click.option(
    '-x', '--dryrun', 'dryrun',
    is_flag=True, default=False,
    help='Don\'t perform actual sorting'
)
@click.option(
    '-c', '--config', 'config_file',
    envvar='mediasorter_CONFIG',
    type=click.Path(), required=False,
    help="Override default configuration file path."
)
@click.option(
    '-v', '--verbose', 'verbose',
    is_flag=True, default=False,
    help="Show INFO messages."
)
@click.option(
    '-vv', 'extra_verbose',
    is_flag=True, default=False,
    help="Show DEBUG messages."
)
@click.option(
    '-q', '--quiet', 'quiet',
    is_flag=True, default=False,
    help="No console ouput. (!) Performs sorting operations without asking."
)
@click.option(
    '-y', '--yes', 'yes',
    is_flag=True, default=False,
    help="Don't ask for confirmation."
)
@click.option(
    '-l', '--logfile', required=False,
    help="Log to file (overrides the configuration option)."
)
@click.option(
    '--loglevel', required=False, default="INFO",
    help="Specify log level for the file handler (overrides the configuration option)."
)
@click.option(
    "--setup", "setup", is_flag=True, default=False,
    help="Install the sample configuration for the current user."
)
@click.option('--version', "version", is_flag=True, default=False)
@click.argument('src_paths', required=False, nargs=-1)
def cli_root(
    src_paths, dst_path, mediatype, action, infofile, shasum, chown, user, group, file_mode,
    directory_mode, metainfo_tag, dryrun, config_file, logfile, loglevel,
    verbose, extra_verbose, quiet, yes, overwrite, dst_path_mov, dst_path_tv,
    version, setup
):
    console = Console(quiet=quiet)

    # Set up (restrict) logging to console.
    if verbose:
        console_log_level = logging.INFO
    elif extra_verbose:
        console_log_level = logging.DEBUG
    else:
        console_log_level = logging.CRITICAL

    for h in logging.getLogger().handlers:
        h.setLevel(console_log_level)

    # Print version.
    if version:
        from . import __version__
        console.print(Text(__version__), style="bold green")
        sys.exit(0)

    # Easily install default configuration into the expected location.
    if setup:
        src = pkg_resources.resource_filename(__name__, "mediasorter.sample.yml")
        target = DEFAULT_CONFIG_PATHS[0]

        if not os.path.exists(target) or Confirm.ask(f"File already exists {target}, overwrite?"):
            copyfile(src, target)
            console.print(Text(f"Configuration file copied to {target}", style="green bold"))

        api_key = Prompt.ask("Enter TMDB API key", default="skip")
        if api_key == "skip":
            sys.exit(0)

        lines = []
        with open(target) as cfg_file:
            for line in cfg_file:
                lines.append(
                    re.sub(r"# key.*", f"key: \"{api_key}\"", line)
                )
        with open(target, "w") as cfg_file:
            for line in lines:
                cfg_file.write(line)
        sys.exit(0)

    # Load configuration.
    config = _load_config(
        console=console,
        config_files=[config_file] if config_file else DEFAULT_CONFIG_PATHS,
        verbose=extra_verbose
    )
    if not config:
        console.print(Text("Consider using --setup to install default configuration file"))
        console.print(Text("No configuration, use -vv for traceback.", style="red"))
        sys.exit(1)

    # Set up logging to a file.
    if log_file := logfile or (config.loging and config.loging.logfile):
        log.debug(f"Setting up logging to '{log_file}'")
        try:
            file_handler = logging.FileHandler(log_file)
            file_handler.setFormatter(
                logging.Formatter("%(asctime)s [%(module)s] [%(levelname)s] %(message)s")
            )
            file_handler.setLevel(loglevel or config.loging.loglevel or 'INFO')
            logging.getLogger().addHandler(file_handler)
        except PermissionError:
            console.print(Text(f"Can't log to file '{log_file}', permission denied.", style="red"))

    # CLI options override pre-configured values.
    if metainfo_tag is not None:
        config.parameters.movie.allow_metadata_tagging = metainfo_tag

    # Sort TV and movie medias to their respective directories if corresponding options
    # are specified ('dst_path_tv' and 'dst_path_mov' respectively) and if 'media type'
    # allows it. Otherwise, use the same path (dst_path) for both media types.
    dst_path = os.path.abspath(os.path.expanduser(dst_path))
    dst_path_tv = os.path.abspath(os.path.expanduser(dst_path_tv)) if dst_path_tv else dst_path
    dst_path_mov = os.path.abspath(os.path.expanduser(dst_path_mov)) if dst_path_mov else dst_path

    # CLI scan options override pre-configured values.
    cli_options = OperationOptions(
        user=user, group=group, chown=chown, dir_mode=directory_mode, file_mode=file_mode,
        overwrite=overwrite, infofile=infofile, shasum=shasum
    )
    scans = [
        ScanConfig(
            src_path=os.path.abspath(os.path.expanduser(src_path)),
            media_type=MediaType(mediatype),
            action=action,
            tv_shows_output=dst_path_tv if mediatype in ("tv", "auto") else None,
            movies_output=dst_path_mov if mediatype in ("movie", "auto") else None,
            options=cli_options
        ) for src_path in src_paths
    ]

    if scans:
        config.scan_sources = scans
    elif not config.scan_sources:
        console.print(Text("No scans requested.", style="bold red"))
        log.error("No scans requested.")
        sys.exit(1)

    sorter = MediaSorter(config)

    # Crawl through the source path and grab available sorting operations.
    s = console.status(Text("Scanning", style="green bold"))
    if not verbose and not extra_verbose:
        s.start()

    ops = asyncio.run(sorter.scan_all())

    s.stop()

    # Print pre-sort summary.
    console.print()
    console.rule(style="green")

    ops = sorted(ops, key=lambda o: o.is_error, reverse=False)
    for sort_operation in ops:
        _pretty_print_operation(sort_operation, console, before=True)

    to_be_sorted = [o for o in ops if not o.is_error]
    errored = [o for o in ops if o.is_error]

    console.print(
        Text(f"\nOK: {len(to_be_sorted)}", style="green"),
        Text(","),
        Text(f"SKIP: {len(errored)}", style="red" if errored else "green"),
    )

    console.rule(style="red" if any([op.is_error for op in ops]) else "green")

    if not to_be_sorted:
        if errored:
            log.error(f"No valid files for sorting in, {len(errored)} errors.")
            console.print(Text("Nothing to sort!", style="bold red"))
            sys.exit(1)
        else:
            log.info(f"Nothing to sort.")
            console.print(Text("Nothing to sort.", style="green bold"))
            sys.exit(0)

    if dryrun:
        log.info("--dryrun, we're done here.")
        sys.exit(0)

    confirmed = yes or quiet or Confirm.ask("Continue?")

    if not confirmed:
        sys.exit(1)

    # SORT!
    text = TextColumn("{task.description}", table_column=Column(ratio=2))
    bar = BarColumn(table_column=Column(ratio=1))
    with Progress(bar, text, expand=True, transient=True) as progress:
        task = progress.add_task("Sorting", total=len(ops), visible=not quiet)

        for operation in to_be_sorted:
            progress.update(task, description=os.path.basename(operation.output_path))
            try:
                asyncio.run(operation.handler.commit())
            except Exception as e:
                if extra_verbose:
                    log.exception(e)
                console.print(f"{e}", style="red bold")
            progress.update(task, advance=1)

    for sort_operation in to_be_sorted:
        _pretty_print_operation(sort_operation, console)

    # Return non-zero if any of the confirmed sort operations fails.
    if any([op.is_error for op in to_be_sorted]):
        sys.exit(1)


def main():
    cli_root(obj={})


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['mediasorter']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=6.0,<7.0',
 'aiohttp>=3.8.3,<4.0.0',
 'async-cache>=1.1.1,<2.0.0',
 'click>=8.1.3,<9.0.0',
 'pydantic>=1.10.4,<2.0.0',
 'rich>=13.1.0,<14.0.0']

entry_points = \
{'console_scripts': ['mediasorter = mediasorter.cli:main']}

setup_kwargs = {
    'name': 'multimediasorter',
    'version': '1.0.12',
    'description': '',
    'long_description': '# (multi)mediasorter\n\nmediasorter is a tool to automatically "sort" media files from a source naming format\ninto something nicer for humans to read/organize, and for tools like Jellyfin to parse\nand collect metadata for. It uses The Movie DB for movie metadata and TVMaze for\nTV metadata to obtain additional information, then performs the "sort" via\na user-selectable mechanism. In this aspect it seeks to be a replacement for\nFileBot and other similar tools.\n\nMost aspects of mediasorter can be configured, either inside the main configuration file,\nor via command-line arguments; it hopes to remain simple yet flexible, doing exactly\nwhat the administrator wants and nothing more.\n\nmediasorter is free software, released under the GNU GPL version 3 (or later).\nCore of the mediasorter is written in Python 3 using asyncio and a simple CLI script and makes use of Click (`python3-click`) and YAML (`python3-yaml`).\n\n## Usage\n\n1. Install on your machine or in a virtualenv.\n\n``` bash\n$ pip install multimediasorter  # install\n```\n\n``` bash\n# Install the bundled config.yaml that should include all that is needed,\n# (!) except for the TMDB api key.\n$ mediasorter --setup\n```\n\n2. Edit the configuration file with your TMDB API key (otherwise only TV shows searches will work).\n3. Run `mediasorter.py --help` for detailed help.\n\n```bash\n# Or e.g.:\n$ mediasorter tests/test_data/ -dtv ~/Media/Series -dmov ~/Media/Movies\n```\n4. Profit!\n\n## Metainfo Tagging\n\nWith the `-tm`/`--tag-metainfo` option, additional info can be added to the destination filename to leverage Jellyfin\'s ["multiple movie versions"](https://jellyfin.org/docs/general/server/media/movies.html#multiple-versions-of-a-movie) feature. Currently, this only works with Movies (not TV episodes) in Jellyfin, and thus in mediasorter as well.\n\nWhen this option is specified, the information found in the `metainfo_map` in the configuration file which is present in the source filename will be appended, using the square-brackets format, to the end of the destination filename.\n\nWhen parsing, the list is iterated through in the order specified, and then for each item, the source filename is searched for the relevant regex match. If found, the value will be appended (once) to the metainfo string. The entries are grouped by type, for example cuts/editions first, then resolutions, then media types, etc. to produce a coherent and consistent string.\n\nA large sample of possible match values is included in the `mediasorter.yml.sample` file, but more can be added or some removed as desired.\n\nAs an example, the following might be a destination filename with metainfo tagging using the default map:\n\n```\nLord of the Rings: The Return of the King, The (2003) - [Extended Edition 2160p BD Remux 7.x Atmos TrueHD].mkv\n```\n\n## Replacement\n\nBy default, `mediasorter` will replace an existing destination file, if one exists,\nwith a new one during a run. This is useful if new media comes in which should\nreplace the existing media (e.g. an upgraded quality version). To disable this\nbehaviour, use `--no-upgrade`.\n\nThis behaviour is redundant when meta-info tagging is enabled for Movies, since the\ndifferentiating quality profile would trigger a new file to be created anyway; it is\nthus mostly useful for TV which does not support this feature.\n\n## Search Overrides\n\nSometimes, the name of a piece of media, as extracted from the file, will not return\nproper results from the upstream metadata providers. If this happens, `mediasorter`\nincludes an option in the configuration file to specify search overrides.\nFor example, the TV show "S.W.A.T." does not return sensible results, so it\ncan be overridden like so:\n\n``` yaml\ntv:\n  search_overrides:\n    "s w a t": "swat"\n    # ...\n```\n\nThis is currently the only *provided* example for demonstration purposes,\nbut it can happen to many titles. If you find a title that returns\nno results consider adding it to this list on your local system.\n\n## Name Overrides\n\nSometimes, the name returned by the metadata providers might not\nmatch what you want to sort as. Thus `mediasorter` can override\ntitles based on a list provided in the configuration file. For example,\nif you want the TV show "Star Trek" to be named\n"Star Trek: The Original Series" instead, it can be overridden like so:\n\n``` yaml\ntv:\n  name_overrides:\n    "Star Trek": "Star Trek: The Original Series"\n    # ...\n```\n\nThese overrides are specific to media type (`tv` or `movie`) to avoid conflicts,\ne.g. in this example, with the 2009 film "Star Trek" which would also be changed\n(erroneously) if this were global.\n\nName overrides are handled *before* adjusting a suffixed "The", so entries containing\n"The" should be written normally, e.g. "The Series" instead of "Series, The"\neven if the latter is what is ultimately written.\n\n## Scan multiple directories\n\n`mediasorter` can be asked to scan multiple directories. Either via CLI or via\nthe configuration file - especially handy when running `mediasorter` as a cron job.\n```yaml\n# Use this to configure what directories should be sorted instead of the CLI argument(s).\nscan_sources:\n\n  - src_path: ~/Downloads-01\n    media_type: auto  # force only a specific media type tv/movie/auto\n    tv_shows_output: ~/Media/TV  # where to put recognized TV shows\n    movies_output: ~/Media/Movies\n\n  - src_path: ~/Downloads-02\n    media_type: auto\n    tv_shows_output: ~/Media/TV\n    movies_output: ~/Media/Movies\n```\n\n```bash\n# Crontab\n$ * * * * * mediasorter -a move\n```\n\n## fix-episodes.sh\n\nmediasorter isn\'t that smart. For instance, if a show has inconsistent episode numbers between, say, airdate and a DVD, it can give episodes the wrong numbering.\n\nFixing this manually is quite cumbersome, and after having to deal with it more than once, I created this quick-and-dirty script that will quickly rename such files, especially for ranges of episodes that are incorrectly numbered.\n\nRun it with no arguments for usage information.\n',
    'author': 'None',
    'author_email': 'None',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8',
}


setup(**setup_kwargs)



import logging
import os
import re
from copy import copy
from typing import List, Tuple, Optional, Dict


class ParsingError(Exception):
    pass


# Try all of these to look for a release year in the media name.
# Year is 4 numbers after all, so this can get pretty vague,
# so let's start with the most likely patterns, this is the best we can do (?)
YEAR_PATTERNS = (
    re.compile(r"\(([0-9]{4})\)"),  # (<year>)
    re.compile(r" \(?([0-9]{4})\) "),  # [space]<year>[space]
    re.compile(r" \(?([0-9]{4})\)?$"),  # [space]<year>[end]
    re.compile(r"[^0-9]\(?([0-9]{4})\)?[^0-9]"),  # [non-number]<year>[non-number]
)

# Try all of these patterns to look for Sxx Eyy identifiers.
# Incoming basename will be split and rejoined with spaces, so there is less possibilities...
SxxEyy_PATTERNS = (
    re.compile(r"[Ss]([0-9]+) ?[Ee]([0-9]+)"),  # Show S05 E12
    re.compile(r" (\d{2,3}) (\d{1,3}) "),  # Show 05 24.avi.
    re.compile(r" (\d{1,3})[x ](\d{1,3}) "),  # Show 5[x ]24.avi.
    re.compile(r" E([0-9]+) "),  # Episode only - if surrounded with spaces - we can be quite sure.
    re.compile(r" E([0-9]+)$")  # Same as above, but at the end.
)


# To be used only if tv_show type is forced.
SxxEyy_PATTERNS_EXTENDED = (
    *SxxEyy_PATTERNS,
    re.compile(r"E([0-9][0-9])[^0-9]"),  # Show with episode id only
    re.compile(r"^([0-9][0-9])[^0-9]"),
    re.compile(r"[^0-9]([0-9][0-9])$"),
    re.compile(r"[ \-_]([0-9][0-9])[ \-_]"),
)


# Try all of these patterns to clean up any leftover "junk" in the media name after
# "Sxx Eyy" or "year" data has been extracted.
CLEAN_PATTERNS = (
    re.compile(r"\[[^ ]*]"),
    *YEAR_PATTERNS
)


log = logging.getLogger(".".join([__package__, __name__]))


def split_basename(
        src_path: str,
        split_characters: List[str] = (".", " "),
        min_split_length: int = 3,
        invalid_single_characters: List[str] = ("-",)
) -> List[List[str]]:
    """
    Split a source path file name (basename) into individual words.

    :param src_path:str: Pass the source path to the function
    :param split_characters:List[str]: Specify the characters that are used to split the filename
    :param min_split_length:int: Specify the minimum number of parts that is allowed in the filename
    :param invalid_single_characters:List[str]: Exclude invalid single characters.
    :return: A list of strings
    """
    basename = os.path.basename(src_path)  # Get basename from the path
    filename = os.path.splitext(basename)[0]  # Discard the extension to get the filename

    # Try splitting the filename
    splits = [
        [word for word in filename.split(split_character) if word]  # Discard 'empty' chars
        for split_character in split_characters
    ]
    splits.sort(key=lambda parts: len(parts), reverse=True)
    # filename_parts = splits[0]
    # if len(filename_parts) < min_split_length:
    #     raise ParsingError(
    #         f"Filename '{filename}' could not be split into sufficient parts to be parsed."
    #     )

    result = []
    for split in splits:
        if len(split) < min_split_length:
            continue
        # Some 'special' characters can mess up the search query.
        # Ignore in case of such filename formatting, e.g.: "Westworld - S03E08.avi"
        result.append(  # (!) make it a tuple, so that we can discard duplicates
            tuple(part for part in split if part not in invalid_single_characters)
        )

    if not result:
        raise ParsingError(
            f"Filename '{filename}' could not be split into sufficient parts to be parsed."
        )

    result = [list(split) for split in set(result)]

    return result


def _find_sxx_eyy_in_dis_structure(src_path):
    """
    Try to find "<TV show>/Season <XX>/..."
    """
    season_id = None
    show_name = None

    head, season_dir = os.path.split(os.path.dirname(src_path))
    season_pattern = re.compile(r"[Ss]eason ?(\d+)")
    if match := re.match(season_pattern, season_dir):
        season_id = int(match.groups()[0])
        _, show_name = os.path.split(head)
    return show_name, season_id


def _find_sxx_eyy(
        src_path: str, split_characters: List[str], min_split_length: int, force: bool = False
) -> Tuple[List[str], int, int]:

    show_name, season_id = _find_sxx_eyy_in_dis_structure(src_path)
    if season_id:
        force = True  # We are pretty sure it is a TV show now.

    # Try all the possibilities base on the configured split chars.
    splits = split_basename(src_path, split_characters=split_characters, min_split_length=min_split_length)

    # Start with the longest split.
    splits.sort(key=lambda array: len(array), reverse=True)

    for possible_split in splits:
        string = " ".join(possible_split)
        for pat in SxxEyy_PATTERNS if not force else SxxEyy_PATTERNS_EXTENDED:
            if match := re.search(pat, string):
                # Yay! That was easy...
                if len(match.groups()) == 1:
                    season_id = str(1) if not season_id else season_id
                    episode_id = match.groups()[0]
                else:
                    season_id, episode_id = match.groups()

                season_id = int(season_id)
                episode_id = int(episode_id)
                start_index, end_index = match.span()

                if not show_name:
                    show_name = string[:start_index]
                if not show_name:
                    show_name = string[end_index:]

                return show_name.split(), season_id, episode_id
    raise ParsingError(f"No valid Sxx Eyy found in '{src_path}'.")


def _find_title_and_year(
        src_path: str,
        split_characters: List[str],
        min_split_length: int,
        metadata_mapping: Dict[str, str] = None
) -> Tuple[List[str], Optional[int], List[str]]:
    """
    The _find_title_and_year function takes a path to a file and splits it into its title and year.

    :param src_path: Get the source path of a file
    :param split_characters: Define the characters that will be used to split the string
    :param min_split_length: Determine how many characters we should split on
    :param metadata_mapping: Specify metadata mapping to be extracted.
    :return: A tuple of two values:
    :doc-author: Trelent
    """
    metadata_mapping = metadata_mapping or {}

    # Try all the possibilities based on the configured split chars.
    splits = split_basename(src_path, split_characters=split_characters, min_split_length=min_split_length)

    # Start with the longest split.
    splits.sort(key=lambda array: len(array), reverse=True)
    if len(splits) > 1:
        # If we have more than a single split, discard the 'single word' ones.
        splits = [split for split in splits if len(split) > 1]

    # Go by patterns - there are many possibilities, we want to try to find the best patterns first.
    # split, year, extracted meatainfo
    parsed_result = [], None, []
    probable_result = [], None, []
    for pat in YEAR_PATTERNS:
        for possible_split in splits:
            parsed_metainfo = []
            # Ignore metadata "pieces", some can get confused as year.
            cleaned_split = copy(possible_split)
            for md_key, value in metadata_mapping.items():
                for word in possible_split:
                    if re.fullmatch(md_key, word) and value not in parsed_metainfo:
                        parsed_metainfo.append(value)
                        cleaned_split.remove(word)
            selected_split = cleaned_split

            string = " ".join(selected_split)
            if match := re.search(pat, string):
                matched_group = match.groups()[-1]
                start_index = match.span()[-2]
                selected_split = string[:start_index].split()
                parsed_year = int(matched_group)
                return selected_split, parsed_year, parsed_metainfo
            # Any metadata tags recognized? This could be the right result
            if len(probable_result[2]) < len(parsed_metainfo):
                probable_result = selected_split, None, parsed_metainfo
                continue
            # No other success indicator, just grab the longest split and hope for the best.
            if len(probable_result[0]) < len(selected_split):
                probable_result = selected_split, None, parsed_metainfo
    else:
        # Year is NOT present in the file name.
        log.debug(f"Can't find 'year' in src file: '{src_path}', best guess: {probable_result}")
        parsed_result = probable_result

    return parsed_result


def parse_season_and_episode(
        src_path: str, split_characters: List[str], min_split_length: int, force: bool = False
) -> Optional[Tuple[str, int, int]]:
    """
    The parse_season_and_episode function attempts to parse a series name, season number, and episode number from the
    given source path. It does this by first splitting the filename into words (separated by spaces), then searching for
    words that match SxxExx in them. If one is found, it is assumed to be part of the series title and removed from the list
    of words. The remaining words are then searched for numbers that could be either season or episode identifiers; if both
    are found, they are returned as a tuple along with the cleaned-up version of what was originally in src_path.

    :param src_path:str: Specify the path to the file that is being processed
    :param split_characters:List[str]: Specify a list of characters that should be used to split the filename into parts
    :param min_split_length:int: Specify the minimum length of a split word to be considered
    :param force:bool=False: Force the function to search for sxxexx even if it is not found in the filename
    :return: A tuple of three values:
    :doc-author: Trelent
    """
    """Try to search for and parse series and episode identifiers (SxxEyy)."""
    filename_parts, season_id, episode_id = _find_sxx_eyy(
        src_path, split_characters=split_characters, min_split_length=min_split_length, force=force
    )

    raw_series_title = list()
    for word in filename_parts:

        # Skip years in the title, because of The Grand Tour
        for pat in YEAR_PATTERNS:
            if re.search(pat, word):
                break
        else:
            raw_series_title.append(word)

    final_name = ' '.join([x.lower() for x in raw_series_title])
    for pat in CLEAN_PATTERNS:
        final_name = re.sub(pat, "", final_name).strip()

    return final_name, season_id, episode_id


def parse_movie_name(
        src_path: str,
        split_characters: List[str],
        min_split_length: int,
        metadata_mapping: Dict[str, str] = None
) -> Tuple[str, Optional[int], List[str]]:
    """Try to search for and parse movie title and release year."""
    # Pick the longest (= best chance of the right one in case of a mixed name).
    # filename_parts = split_basename(src_path, split_characters, min_split_length)[0]

    filename_parts, movie_year, metainfo_map = _find_title_and_year(
        src_path,
        split_characters=split_characters,
        min_split_length=min_split_length,
        metadata_mapping=metadata_mapping
    )

    if not movie_year:
        log.warning(f"No valid year found in '{src_path}'.")

    final_name = " ".join([part.lower() for part in filename_parts])
    for pat in CLEAN_PATTERNS:
        final_name = re.sub(pat, "", final_name)

    return final_name, movie_year, metainfo_map


def fix_leading_the(series_title):
    """Fix leading The's in the series title"""
    if re.match('[Tt]he\s(.*)', series_title):
        return re.match('[Tt]he\s(.*)', series_title).group(1) + ', The'
    return series_title



"""
MediaSorter - Sort a media file into an organized library

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""
import asyncio
import logging
import os
import subprocess
from itertools import chain
from typing import Optional, List, Union, Tuple, Type, Any

from pydantic import BaseModel

from .config import (
    MediaSorterConfig,
    ScanConfig,
    MetadataProviderApi,
    OperationOptions,
    MediaType,
    Action
)
from .metadata import (
    TvShowMetadata,
    MovieMetadata,
    MetadataApi,
    TvShowMetadataApi,
    MovieMetadataApi,
    MetadataQueryError,
    MetadataProvider,
)
from .parser import (
    parse_season_and_episode,
    parse_movie_name,
    fix_leading_the,
    ParsingError
)
from .runner import ExecutionError, Executable

log = logging.getLogger(".".join([__package__, __name__]))


class MediaSorterError(Exception):
    pass


class CantSortError(MediaSorterError):
    pass


class Operation(BaseModel):
    input_path: str
    output_path: Optional[str]
    action: Action = Action.COPY
    type: MediaType = MediaType.AUTO
    exception: Optional[Any]
    options: Optional[OperationOptions]

    @property
    def is_error(self):
        return self.exception is not None

    @property
    def handler(self):
        return OperationHandler(self, options=self.options)

    def raise_error(self):
        if self.exception:
            raise self.exception
        return self


class OperationHandler:

    op: Operation
    options: OperationOptions

    def __init__(self, operation, options=None) -> None:
        self.op = operation
        self.options = options or OperationOptions()

    def pre_commit(self):
        if not self.op.output_path:
            self.op.exception = CantSortError(f"Destination path missing.")

        if os.path.exists(self.op.output_path):
            log.info(f"File exists '{self.op.output_path}'")
            if self.options.overwrite:
                log.info(f"Removing for overwrite.")
                os.remove(self.op.output_path)
            else:
                msg = f"Destination file '{self.op.output_path}' exists, overwrite not allowed."
                log.warning(msg)
                self.op.exception = CantSortError(msg)

    async def commit(self):
        self.pre_commit()
        if self.op.is_error:
            return

        try:
            Executable.from_action_type(self.op.action) \
                      .commit(self.op.input_path, self.op.output_path)

            uid, gid = None, None
            if self.options.chown:
                uid, gid = _get_uid_and_gid(self.options.user, self.options.group)
                log.info(
                    f"Correcting ownership and permissions: "
                    f"{uid=}, {gid=}, mode={self.options.file_mode}"
                )
                parent_dir = os.path.dirname(self.op.output_path)
                os.chown(parent_dir, uid, gid)
                os.chown(self.op.output_path, uid, gid)
                os.chmod(self.op.output_path, int(self.options.file_mode, 8))
                if self.options.dir_mode:
                    log.info(f"Changing parent dire mode: {self.options.dir_mode=}")
                    os.chmod(parent_dir, int(self.options.dir_mode, 8))

            # Create the info file.
            if self.options.infofile:
                info_file_name = f"{self.op.output_path}.txt"
                log.info(f"Creating info file: .../{os.path.basename(info_file_name)}")
                info_file_contents = [
                    "Source filename:  {}".format(os.path.basename(self.op.output_path)),
                    "Source directory: {}".format(os.path.dirname(self.op.output_path))
                ]
                with open(info_file_name, 'w') as fh:
                    fh.write('\n'.join(info_file_contents))
                    fh.write('\n')
                if self.options.chown:
                    os.chown(info_file_name, uid, gid)
                    os.chmod(info_file_name, int(self.options.file_mode, 8))

            # Create sha256sum file
            if self.options.shasum:
                shasum_name = '{}.sha256sum'.format(self.op.output_path)
                log.debug(f"Generating shasum file: .../'{os.path.basename(shasum_name)}'.")
                shasum_cmdout = subprocess.run(
                    ['sha256sum', '-b', f'{self.op.output_path}'],
                    capture_output=True, encoding='utf8'
                )
                if shasum_cmdout.returncode != 0 or not shasum_cmdout.stdout:
                    msg = f"SHASUM checksum generation failed, " \
                          f"out={shasum_cmdout.stdout} err={shasum_cmdout.stderr}"
                    log.error(msg)
                    self.op.exception = MediaSorterError(msg)
                    return

                shasum_data = shasum_cmdout.stdout.strip()
                log.info(
                    f".../{os.path.basename(self.op.output_path)}: SHA generated {shasum_data}.")
                with open(shasum_name, 'w') as fh:
                    fh.write(shasum_data)
                    fh.write('\n')
                if self.options.chown:
                    log.debug(f"{os.path.basename(shasum_name)}: changing owner.")
                    os.chown(shasum_name, uid, gid)
                    os.chmod(shasum_name, int(self.options.file_mode, 8))

        except ExecutionError as e:
            log.error(f"Unexpected error: {e}")
            self.op.exception = e
            return


def _get_uid_and_gid(user_name=None, group_name=None):
    # expect ImportError on Windows
    import grp
    import pwd

    if user_name:
        uid = pwd.getpwnam(user_name)[2]
    else:
        uid = os.getuid()

    if group_name:
        gid = grp.getgrnam(group_name)[2]
    else:
        gid = os.getgid()

    return uid, gid


class MediaSorter:

    config: MediaSorterConfig

    def __init__(self, config) -> None:
        self.config = config

    @classmethod
    def from_src_path(
        cls,
        config: MediaSorterConfig,
        src_path: Union[List[str], str],
        tv_shows_output: str,
        movies_output: str,
        action: Action = Action.COPY,
        media_type: MediaType = MediaType.AUTO,
        options: OperationOptions = None
    ):
        # Construct a 'config' object from CLI parameters.
        src_paths = [src_path] if isinstance(src_path, str) else src_path
        scans = [
            ScanConfig(
                src_path=src_path,
                tv_shows_output=tv_shows_output,
                movies_output=movies_output,
                action=action,
                media_type=media_type,
                options=options or OperationOptions()
            ) for src_path in src_paths
        ]
        config.scan_sources = scans
        return cls(config)

    async def scan_all(self) -> List[Operation]:
        """Scan all preconfigured scan sources."""
        scan_ops = [self.scan(**scan.__dict__) for scan in self.config.scan_sources]
        result_lists = await asyncio.gather(*scan_ops)
        return list(chain(*result_lists))

    async def scan(
            self,
            src_path: str,
            media_type: MediaType,
            tv_shows_output: str = None,
            movies_output: str = None,
            action: Action = Action.COPY,
            options: OperationOptions = OperationOptions()
    ) -> List[Operation]:
        """Scan a single source path (file or directory)."""
        operations = []

        if os.path.isdir(src_path):
            tasks = []
            log.debug(f"Scanning {src_path} [{media_type}]")
            for filename in sorted(os.listdir(src_path)):
                child_path = os.path.join(src_path, filename)
                tasks.append(
                    self.scan(
                        child_path,media_type, tv_shows_output,
                        movies_output, action, options
                    )
                )

            results = await asyncio.gather(*tasks)
            for res in results:
                operations.extend(res)
        elif not os.path.exists(src_path):
            log.error(f"{src_path}: path does not exist!")
            op = Operation(input_path=src_path)
            op.exception = FileNotFoundError(f"File does not exist: '{src_path}'")
            if options:
                op.options = options
            operations.append(op)
        else:
            if op := await self.suggest(src_path, media_type=media_type, action=action):
                if op.is_error:
                    pass
                elif op.type == MediaType.TV_SHOW and tv_shows_output:
                    op.output_path = os.path.join(tv_shows_output, op.output_path)
                elif op.type == MediaType.MOVIE and movies_output:
                    op.output_path = os.path.join(movies_output, op.output_path)
                if options:
                    op.options = options
                operations.append(op)

        return [op for op in operations if op]

    @staticmethod
    def _get_api(api: MetadataProviderApi) -> Optional[MetadataApi]:
        """API instantiation and null check for the Enum(<str>)."""
        try:
            return MetadataProvider(api.name).clazz(api)
        except ValueError:
            log.warning(
                f"'{api.name}' API provider "
                f"not recognized among {[e.value for e in MetadataProvider]}"
            )

    def _get_apis(self, type_: Type[MetadataApi]) -> List[MetadataApi]:
        """Get all available MD providers by type."""
        return list(
            filter(
                lambda x: x is not None, filter(
                    lambda provider: isinstance(provider, type_), map(
                        lambda api: self._get_api(api), self.config.api)
                )
            )
        )

    async def _query(
        self, api_type: Type[MetadataApi], *args
    ) -> Union[TvShowMetadata, MovieMetadata]:
        """Make an external query to find a TV-show/movie metadata."""
        apis = self._get_apis(api_type)
        if not apis:
            msg = f"No '{api_type.__name__}' metadata provider APIs configured."
            log.error(msg)
            raise MediaSorterError(msg)

        # Try to fetch info from all available APIs at once, return only the first result
        api_queries = [asyncio.create_task(api.query(*args)) for api in apis]
        exceptions = []
        for api_query in asyncio.as_completed(api_queries):
            try:
                first_result = await api_query
            except MetadataQueryError as e:
                log.warning(str(e))
                exceptions.append(e)
                continue
            if first_result:
                return first_result

        exceptions.append(
            MediaSorterError(
                f"{api_type.__name__}: none of the "
                f"{[a.__class__.__name__ for a in apis]} APIs was successful"
            )
        )
        raise MediaSorterError(exceptions)

    async def suggest_tv_show(self, src_path: str):
        parsed_tv_show = parse_season_and_episode(
            src_path,
            self.config.parameters.split_characters,
            self.config.parameters.tv.min_split_length,
            force=True  # Try everything!
        )

        if parsed_tv_show:
            name, series, episode = parsed_tv_show

            log.debug(f"TV show recognized: series='{name}' S={series} E={episode}")
            result = await self._query(TvShowMetadataApi, name, series, episode)

            if self.config.parameters.tv.suffix_the:
                result.series_title = fix_leading_the(result.series_title)

            # Build the final path+filename
            season_dir = self.config.parameters.tv.dir_format.format(**result.__dict__)
            filename = self.config.parameters.tv.file_format.format(**result.__dict__)
            filename = " ".join(filename.split())

            return season_dir, filename

    async def suggest_movie(self, src_path: str) -> Tuple[Optional[str], str]:
        """
        Suggest the title of the movie, as well as its year based on an external metadata API.

        :param src_path: str: Specify the path to the file that is going to be moved
        :return: A final, formatted movie file name suggestion (dir and filename)
        """
        try:
            # Even if movie type is forced, try to find the season/episode numbers
            # to disqualify the media file before any network requests.
            if parse_season_and_episode(
                    src_path,
                    self.config.parameters.split_characters,
                    self.config.parameters.movie.min_split_length,
                    force=False  # We DON'T want to parse a TV show at all costs.
            ):
                raise MediaSorterError(f"This appears to be a TV show: {src_path}")
        except ParsingError:
            pass

        movie, year, metainfo = parse_movie_name(
            src_path,
            self.config.parameters.split_characters,
            self.config.parameters.movie.min_split_length,
            self.config.metainfo_map
        )
        log.debug(f"Parsed {os.path.basename(src_path)}, {movie=} {year=}")
        result = await self._query(MovieMetadataApi, movie, year)

        for title in self.config.parameters.movie.name_overrides:
            if title == result.title:
                result.title = self.config.parameters.movie.name_overrides[title]
                break

        filename = self.config.parameters.movie.file_format.format(**result.__dict__)

        # Sort movie files in a directory.
        if self.config.parameters.movie.subdir:
            subdir = self.config.parameters.movie.dir_format.format(**result.__dict__)
        else:
            subdir = None

        if self.config.parameters.movie.allow_metadata_tagging and metainfo:
            # MUST be "space", "hyphen", "space"
            # https://jellyfin.org/docs/general/server/media/movies/#multiple-versions-of-a-movie
            filename = f"{filename} - [{' '.join(metainfo)}]"

        return subdir, filename.strip()

    async def suggest(
            self, src_path: str, media_type: MediaType = MediaType.AUTO, action: Action = Action.COPY
    ) -> Optional[Operation]:

        extension = os.path.splitext(src_path)[-1]

        log.info(f">>> Parsing {src_path} [{media_type}]")

        if not extension:
            log.warning(f"{os.path.basename(src_path)}: files without extension not allowed.")
            return None
        elif extension and extension not in self.config.parameters.valid_extensions:
            log.warning(
                f"{os.path.basename(src_path)}: extension '{extension}' not allowed, "
                f"not in {self.config.parameters.valid_extensions}."
            )
            return None

        # First try to parse a TV show (series and episodes numbers)
        directory, filename = None, None
        operation = Operation(input_path=src_path, type=MediaType.TV_SHOW, action=action)
        if media_type in [MediaType.AUTO, MediaType.TV_SHOW]:
            try:
                directory, filename = await self.suggest_tv_show(src_path)
            except ParsingError as e:
                msg = f"{os.path.basename(src_path)} can't be parsed into a TV show: {e}."
                if media_type == MediaType.TV_SHOW:
                    log.error(msg)
                    operation.exception = MediaSorterError(msg)
                    return operation
                log.debug(msg)
            except (MediaSorterError, MetadataQueryError) as e:
                operation.exception = e
                return operation

        if not filename:
            # Not a TV show? Must be a movie then...
            operation.type = MediaType.MOVIE
            try:
                directory, filename = await self.suggest_movie(src_path)
            except (MediaSorterError, ParsingError) as e:
                msg = f"{os.path.basename(src_path)} can't be parsed into a movie title: {e}."
                log.error(msg)
                operation.exception = MediaSorterError(msg)
                return operation
            except (MediaSorterError, MetadataQueryError) as e:
                operation.exception = e
                return operation

        # Build the final path+filename
        if directory:
            dst_path = os.path.join(directory, filename)
        else:
            dst_path = os.path.join(filename)

        # Get rid of forbidden characters (I'm looking at you, Windows!)
        for illegal_char in (":", "#"):
            dst_path = dst_path.replace(illegal_char, "")

        dst_path += extension
        log.debug(f"Suggested output path: {dst_path}")
        operation.output_path = dst_path

        return operation

    @staticmethod
    async def commit_all(operations: List[Operation]) -> List[Operation]:
        tasks = [sort_operation.handler.commit() for sort_operation in operations]
        return await asyncio.gather(*tasks)
